
from fastcore.script import *
from pathlib import Path
import glob
import os
import numpy as np
from espiownage.core import *
import shutil
#from functools import partial
#import multiprocessing as mp

""" Grabs most recent version of annotations from various "annotations*" directories
"""




@call_parse
def grab_recent(
    dirs:Param("annotation directories check", str)='annotations*',
    dest:Param("Directory to write new annotations to",str)='recent_annotations',
    ):


    dirs = ''.join(dirs)  # convert to str
    print("dirs = ",dirs)
    dir_list = sorted(glob.glob(dirs)) # list of all annotation .csv files for ellipses
    print(dir_list)
    main_dir = dir_list[0]
    meta_files = [os.path.basename(x) for x in sorted(glob.glob(main_dir+'/'+'*.csv'))]
    recent_files = []
    more_recent_than_orig_count = 0
    for file in meta_files:
        time = -20.0
        recent_file = ''
        more_recent = 0
        for i, dir in enumerate(dir_list):
            full_filename = dir+'/'+file
            if os.path.exists(full_filename):
                this_time = os.path.getmtime(full_filename)
                if  this_time > time:
                    time, recent_file = this_time, full_filename
                    if i>0: more_recent = 1
        more_recent_than_orig_count += more_recent
        recent_files.append(recent_file)

    mkdir_if_needed(dest)

    for src_file in recent_files:
        dest_file =  dest+'/'+os.path.basename(src_file)
        print("src_file, dest_file =",src_file, dest_file)
        shutil.copyfile(src_file, dest_file)

    print(f"{len(recent_files)} most recent files. {more_recent_than_orig_count} more recent than original.")

    return



#! /usr/bin/env python3

'''
Ellipse editor, standalone utility for use with steelpan images & annotations
Author: Scott H. Hawley  License: MIT (below)

Reads a CSV file of the format cx,cy,a,b,angle,rings  (a & b are ellipse axes), as in..
$ cat test_img.csv
37,42,30,42,63,0
402,71,37,20,41,0
459,256,40,19,45,0
175,198,41,30,134,0

To get the corresponding image file, it replaces '.csv' with '.png' in the file name

For more options, run $ ./ellipse_editor.py --help


This code actually uses Tk instead OpenCV.  Because...those are the first tutorials I found.
The rest of the SPNet code uses OpenCV.



MIT License:
Copyright (c) 2018 Scott Hawley, https://github.com/drscotthawley

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''


import tkinter as tk
from tkinter.simpledialog import askfloat
import math
import numpy as np
import pandas as pd
from PIL import Image, ImageTk, ImageOps
import tkinter.font
import argparse
import os
import glob
import sys
from pathlib import Path
from fastcore.script import *
import re
from espiownage.core import *
from itertools import cycle
from collections import defaultdict



def increment(num, digits,step):
    # return the first part of match as is, Return the 2nd match + 1 which is 'x + 1'
    return num.group(1) + str(int(num.group(2)) + step).zfill(digits)

def get_next_img(meta_file, step=1, img_bank='images/'):
    "Increments the number of the image filename and tries to load it"
    digits = len(Path(meta_file).stem.split('_')[-1]) # num digits in ending number part of the name
    next_meta = re.sub('(proc_)([0-9]{'+str(digits)+'})', lambda m: increment(m,digits,step), meta_file)
    next_img_path = meta_to_img_path(next_meta, img_bank=img_bank)
    if os.path.exists(str(next_img_path)):
        return ImageTk.PhotoImage(image=Image.open(next_img_path)), str(next_img_path)
    blank_img = Image.new('RGB', (512, 384), (255, 255, 255))
    return ImageTk.PhotoImage(image=blank_img), 'None'



def poly_oval(cx, cy, a, b, angle=0, steps=100 ):
    # From https://mail.python.org/pipermail/python-list/2000-December/022013.html
    """return an oval as coordinates suitable for create_polygon"""

    # angle is in degrees anti-clockwise, convert to radians
    rotation = angle * math.pi / 180.0  # overall angle of the ellipse

    point_list = []

    # create the oval as a list of points
    for i in range(steps):

        # Calculate the angle for this step: 360 degrees == 2 pi radians
        theta = (math.pi * 2) * (float(i) / steps)  # theta is for points drawing the circumference of the ellipse
        x1, y1 = a * math.cos(theta), b * math.sin(theta)

        # rotate x, y
        x = (x1 * math.cos(rotation)) + (y1 * math.sin(rotation))
        y = (y1 * math.cos(rotation)) - (x1 * math.sin(rotation))

        point_list.append((x + cx))
        point_list.append((y + cy))

    return point_list


def clean_pandas_list(list_):
    list_ = list_.replace(', ', '","')
    list_ = list_.replace('[', '["')
    list_ = list_.replace(']', '"]')
    return list_


def interleave_lists(list1:list, list2:list):
    "alternate between 1 and 2; if they're of different lengths, it keeps going with whatever's left in the longer one"
    return [x for both in zip(cycle(list1), list2) for x in both]

def dedup_list(li):
    "remove duplicates while preserving order of first dup(non-dup?/orig?).  cf. https://stackoverflow.com/a/480227/4259243"
    seen = set()
    seen_add = seen.add
    return [x for x in li if not (x in seen or seen_add(x))]


def get_top_loss_list(
    top_losses_dir,  # directory where top-losses info is stored
    ):
    "gets list of filenames with top losses for use with preferential ordering"
    if (top_losses_dir==None) or (not os.path.exists(top_losses_dir)): return []
    # slurp in any csv files we've got
    # note they likely have different columns, and the loss values are not comparable
    tldir_files_list = glob.glob(top_losses_dir+'/*.csv') # files in tl directory
    top_loss_list = []
    for tldir_file in tldir_files_list:
        #print("top losses tldir_file = ",tldir_file)
        df = pd.read_csv(tldir_file)
        fnames = dedup_list(list(df["filename"]))
        if [] == top_loss_list: top_loss_list = fnames
        else: top_loss_list = interleave_lists(top_loss_list, fnames)
    return dedup_list(top_loss_list)


class EllipseEditor(tk.Frame):
    '''Edit ellipses for steelpan images'''

    def __init__(self, parent,  # tk class and parent window
        meta_file_list,         # list of csv files to edi
        img_bank='images/',     # where images are stored
        tldir=None,        # directory where top-losses info is stored
        seq=True,               # ignore top losses and do sequential selection of frames (per existing annotations)
        ):
        tk.Frame.__init__(self, parent)
        self.meta_file_list = meta_file_list
        self.img_bank = img_bank
        self.top_loss_list = get_top_loss_list(tldir)
        if self.top_loss_list == []: seq = True
        if not seq: self.meta_file_list = combine_file_and_tl_lists(self.meta_file_list, self.top_loss_list)
        self.seq = seq

        # create a canvas
        self.width, self.height = 512, 384   # size of images
        self.y0 = 0# self.height    # y offset for all operations
        self.readout = 700                   # width for additional annotation text
        self.canvas = tk.Canvas(width=self.width + self.readout, height=20+2*self.height )
        self.canvas.pack(fill="both", expand=True)

        self.file_index = 0                 # TODO: change to grab from top losses
        self.meta_file = meta_file_list[self.file_index]
        self.img_file = str(meta_to_img_path(self.meta_file, img_bank=self.img_bank))

        self.mask_pred_file = ''
        self.mask_img = None
        self.showing_mask = True

        self.bbox_pred_file = 'top_losses/bboxes_top_losses_real.csv' # TODO: this is fragile
        self.showing_bboxes = True
        self.bbox_list = []
        self.bbox_df = pd.read_csv(self.bbox_pred_file, converters={'bblist': eval})
        self.bbox_df.apply(clean_pandas_list)

        self.tl_ring_count_file, self.tl_ring_count_dict = '', {}
        self.tl_ring_count_dict = defaultdict(lambda: [], self.tl_ring_count_dict) # map filenames to lists of rings info; defaults to empty list
        tl_rc_files = glob.glob(tldir+'/*ring*.csv')
        if len(tl_rc_files) > 0:
            self.tl_ring_count_file = tl_rc_files[0] # not sure I'll keep the same name. something ring-related
            self.setup_tl_ring_count_dict()
        self.showing_predrings, self.predringlist = True, []  # thing we actually use

        self.color = "green"

        # this data is used to keep track of an item being dragged
        self._drag_data = {"x": 0, "y": 0, "items": None}

        self._token_data = []
        self._numtokens = 0
        self.hr = 4             # handle radius

        # Define global event bindings
        self.canvas.bind("<B1-Motion>", self.update_readout)
        self.canvas.bind("<Double-Button-1>", self.on_doubleclick)
        self.canvas.bind("<ButtonPress-2>", self.on_rightpress)   # on mac, button 2 is right mouse
        self.canvas.bind("<ButtonPress-3>", self.on_rightpress)   # on linux, button 3 is right mouse

        self.canvas.focus_set()
        self.canvas.bind("<Q>", self.on_qkey)
        self.canvas.bind("<q>", self.on_qkey)
        self.canvas.bind("<S>", self.on_skey)
        self.canvas.bind("<s>", self.on_skey)
        self.canvas.bind("<M>", self.on_mkey)
        self.canvas.bind("<m>", self.on_mkey)
        self.canvas.bind("<R>", self.on_rkey)
        self.canvas.bind("<r>", self.on_rkey)
        self.canvas.bind("<B>", self.on_bkey)
        self.canvas.bind("<b>", self.on_bkey)
        self.canvas.bind("<Left>", self.on_leftarrow)
        self.canvas.bind("<Right>", self.on_rightarrow)
        self.canvas.bind('<Motion>', self.mouse_move)

        self.infostr = ""
        self.text = self.canvas.create_text(self.width+10, 10+self.height, text=self.infostr,
            anchor=tk.NW, font=tk.font.Font(size=15,family='Consolas'))
        self.df = ''

        self.load_new_files()


    def setup_tl_ring_count_dict(self):
        "this will store info about all the antinode ring counts for each file, as a list of lists for each file"
        df = pd.read_csv(self.tl_ring_count_file)
        #print(df.head())
        for i, row in df.iterrows(): # stop telling me stop using iterrows, it's clear coding ;-)
            meta, parts = meta_from_str(row['filename']), row['filename'].split('_')[-5:]
            bbox, rings = [int(x) for x in parts[0:4]], float(row['prediction'])
            bb_rings = bbox + [rings]
            if len(self.tl_ring_count_dict[meta]) == 0: self.tl_ring_count_dict[meta] = [bb_rings]
            else: self.tl_ring_count_dict[meta].append(bb_rings)

    def setup_pred_mask(self):
        self.mask_img = None
        self.mask_pred_file = 'top_losses/seg_images/'+str(Path(self.meta_file).stem)+'_pred.png'
        if os.path.exists(self.mask_pred_file):
            self.mask_img = Image.open(self.mask_pred_file)
            if self.mask_img.size != (self.width, self.height): # to allow for half-size masks
                self.mask_img.size = self.mask_img.size.resize((self.width, self.height))
            self.mask_img = ImageOps.colorize(self.mask_img, black ="black", white =(150,0,150))
        return

    def merge_mask_image(self):
        if not self.mask_img: return
        if self.showing_mask:
            self.image = Image.blend(self.image, self.mask_img, 0.5)
            self.assign_image()

    def draw_pred_rings(self):
        if (not self.showing_predrings) or (len(self.predringlist) == 0): return
        for bbr in self.predringlist:
            cx, cy, ringstr = int((bbr[0]+bbr[2])/2), int((bbr[1]+bbr[3])/2), '{0:.1f}'.format(bbr[-1])
            #print("predicted ring counts: ",cx, cy, ringstr)
            ringtext = self.canvas.create_text(cx, self.y0+cy-15, text=ringstr, anchor=tk.CENTER, font=tk.font.Font(size=15), fill="yellow")

    def draw_pred_bboxes(self, please_fix=True):
        if (not self.showing_bboxes) or len(self.bbox_list)==0: return
        for bb in self.bbox_list[0]:
            if please_fix:
                # icevision shrank our images and then ebedded them in 384,384, we need to undo that?
                bb = [int(x*512/384) for x in bb]  # unshrink everything
                bb[1], bb[3] = bb[1]-(512-384)//2, bb[3]-(512-384)//2
            box = self.canvas.create_rectangle(bb[0],bb[1],bb[2],bb[3], outline="cyan", width=2)

    def load_new_files(self):
        self.canvas.delete("all")  #destroy old tokens
        self.text = self.canvas.create_text(self.width+10, 10+self.y0, text=self.infostr,
            anchor=tk.NW, font=tk.font.Font(size=15,family='Consolas'))
        self.meta_file = self.meta_file_list[self.file_index]
        self.img_file = meta_to_img_path(self.meta_file, img_bank=self.img_bank)
        self.setup_pred_mask()
        self.read_assign_image()
        self.merge_mask_image()

        self.read_prev_next_imgs()
        self.bbox_list = self.bbox_df[self.bbox_df['filename'] == os.path.basename(self.meta_file)]['bblist'].tolist() # nice, huh? ;-)
        self.draw_pred_bboxes()
        self.predringlist = self.tl_ring_count_dict[os.path.basename(self.meta_file)]
        self.draw_pred_rings()
        self.read_assign_csv()


    def assign_image(self):
        self.tkimage = ImageTk.PhotoImage(image=self.image)
        self.label = tk.Label(image=self.tkimage)
        self.label.image = self.tkimage # keep a reference!
        self.canvas.create_image(self.width/2, self.y0 + self.height/2, image=self.tkimage)

    def read_assign_image(self):
        self.image = Image.open(self.img_file)
        self.image = ImageOps.colorize(self.image, black ="black", white ="white")
        self.backup = self.image
        self.assign_image()

    def read_assign_csv(self):
        col_names = ['cx', 'cy', 'a', 'b', 'angle', 'rings']
        self.df = pd.read_csv(self.meta_file,header=None,names=col_names)  # read metadata file
        self.df.drop_duplicates(inplace=True)  # sometimes the data from Zooniverse has duplicate rows
        # assign  ellipse tokens (and their handles)
        for index, row in self.df.iterrows() :
            cx, cy, a, b, angle, rings = row['cx'], row['cy'], row['a'], row['b'], float(row['angle']), row['rings']
            a, b, angle = fix_abangle(a,b,angle)
            if (0!=rings):    # 0 rings means no antinode, i.e. nothing there
                self._create_token((cx, cy), (a, b), angle, rings, self.color)
        self.update_readout(None)

    def read_prev_next_imgs(self):
        # prev image
        self.prev_img, name = get_next_img(self.meta_file, -1, img_bank=self.img_bank)
        if self.prev_img: self.canvas.create_image(self.width/2, self.y0 + self.height+20 + self.height/2, image=self.prev_img)
        imlabel = f"Previous Image: {name.split('/')[-1]}"
        self.canvas.create_text(10, self.y0+self.height, text=imlabel, anchor=tk.NW, font=tk.font.Font(size=12), fill='black')

        # next image
        self.next_img, name = get_next_img(self.meta_file, 1, img_bank=self.img_bank)
        if self.next_img: self.canvas.create_image(self.width+20+self.width/2, self.y0 + self.height+20 + self.height/2, image=self.next_img)
        imlabel = f"Next Image: {name.split('/')[-1]}"
        self.canvas.create_text(self.width+30, self.y0+self.height, text=imlabel, anchor=tk.NW, font=tk.font.Font(size=12), fill='black')
        return

    def _create_token(self, coord, axes, angle, rings, color):
        '''Create a tk token at the given coordinate in the given color'''
        self._numtokens += 1
        (x,y) = coord
        (a,b) = axes
        thistag = "token"+str(self._numtokens)   # each token gets its own unique id, plus the whole ellipse gets a 'main' tag
        oval = self.canvas.create_polygon(*tuple(poly_oval(x, self.y0+y, a, b, angle=angle)),outline=color, fill='', width=3, tags=(thistag,"main"))

        # handles for resize / rotation
        h_a_x, h_a_y = x + a*np.cos(np.deg2rad(angle)),  y - a*np.sin(np.deg2rad(angle))
        h_b_x, h_b_y = x + b*np.sin(np.deg2rad(angle)),  y + b*np.cos(np.deg2rad(angle))
        h_a = self.canvas.create_oval(h_a_x-self.hr, self.y0+h_a_y-self.hr, h_a_x+self.hr, self.y0+h_a_y+self.hr, outline=color, fill=color, width=3, tags=(thistag,"handle","axis_a"))
        h_b = self.canvas.create_oval(h_b_x-self.hr, self.y0+h_b_y-self.hr, h_b_x+self.hr, self.y0+h_b_y+self.hr, outline=color, fill="blue", width=3, tags=(thistag,"handle","axis_b"))

        ringstr = '{0:.1f}'.format(rings)
        ringtext = self.canvas.create_text(x-5, self.y0+y-10, text=ringstr, anchor=tk.NW, font=tk.font.Font(size=16), fill=color, tags=(thistag,"ringtext"))

        self._token_data.append([oval,h_a,h_b,ringtext])

        # Define Event Bindings for moving objects around
        self.canvas.tag_bind("main", "<ButtonPress-1>", self.on_main_press)
        self.canvas.tag_bind("main", "<ButtonRelease-1>", self.on_main_release)
        self.canvas.tag_bind("main", "<B1-Motion>", self.on_main_motion)

        self.canvas.tag_bind("handle", "<ButtonPress-1>", self.on_handle_press)
        self.canvas.tag_bind("handle", "<ButtonRelease-1>", self.on_handle_release)
        self.canvas.tag_bind("handle", "<B1-Motion>", self.on_handle_motion)

    def on_qkey(self, event):
        print("Quitting")
        sys.exit()
    def on_skey(self,event):
        print("Saving file ",self.meta_file)
        # TODO: add code to enforce a > b (and fix angle)
        self.df.to_csv(self.meta_file,index=False,header=None)
    def on_mkey(self,event):
        self.showing_mask = not self.showing_mask
        if self.showing_mask and self.mask_img is not None:
            self.merge_mask_image()
            self.load_new_files()
        if not self.showing_mask:
            self.image = self.backup
            self.load_new_files()
    def on_rkey(self,event):
        self.showing_predrings = not self.showing_predrings
        self.load_new_files()
    def on_bkey(self,event):
        self.showing_bboxes = not self.showing_bboxes
        self.load_new_files()
    def on_rightarrow(self,event):               # right arrow on keyboard
        self.file_index += 1                     # TODO: grab from top_losses
        if (self.file_index >= len(self.meta_file_list)):
            self.file_index = 0
        self.load_new_files()
    def on_leftarrow(self,event):
        self.file_index -= 1
        if (self.file_index < 0):
            self.file_index = len(self.meta_file_list)-1
        self.load_new_files()
    def mouse_move(self,event):
        x, y = event.x, event.y
        tx, ty = 2*self.width+40, 2*self.height-40
        box = self.canvas.create_rectangle(tx-5,ty, tx+110,ty+25, fill="white", outline="white")
        self.canvas.create_text(tx, ty, text=f'({x},{y})', anchor=tk.NW, font=tk.font.Font(size=12), fill='black')

    def on_main_press(self, event):
        '''Begining drag of an object'''
        obj_id = self.canvas.find_closest(event.x, event.y)[0]  # record the item and its location
        tags = self.canvas.gettags( obj_id )
        self._drag_data["items"] = tags[0]
        self._drag_data["x"] = event.x
        self._drag_data["y"] = event.y

    def on_main_release(self, event):
        '''End drag of an object'''
        # if object is off the screen, delete it
        if ((event.x < 0) or (event.y < 0 ) or (event.x > self.width) or (event.y > self.height)):
            self.canvas.delete(self._drag_data["items"])
            self.update_readout(None)
        # reset the drag information
        self._drag_data["items"] = None
        self._drag_data["x"] = 0
        self._drag_data["y"] = 0

    def on_main_motion(self, event):
        '''Handle dragging of an object'''
        # compute how much the mouse has moved
        delta_x = event.x - self._drag_data["x"]
        delta_y = event.y - self._drag_data["y"]
        # move the object the appropriate amount
        self.canvas.move(self._drag_data["items"], delta_x, delta_y)
        # record the new position
        self._drag_data["x"] = event.x
        self._drag_data["y"] = event.y



    def retrieve_ellipse_info(self, tokentag):
        # retrieves info for whichever single ellipse is currently being manipulated
        tokenitems = self.canvas.find_withtag( tokentag )
        [main_id, axis_a_id, axis_b_id, ringtext_id ]= tokenitems

        ell_coords = self.canvas.coords(main_id)   # coordinates of all points in ellipse
        cxoords, cyoords = ell_coords[0::2],  ell_coords[1::2]
        cx, cy = np.mean( cxoords ),  np.mean( cyoords )   # coordinates of center of ellipse

        h_a_coords = self.canvas.coords(axis_a_id)
        h_a_x, h_a_y = np.mean( h_a_coords[0::2] ),  np.mean( h_a_coords[1::2] )
        a = np.sqrt( (h_a_x - cx)**2 + (h_a_y - cy)**2 )

        h_b_coords = self.canvas.coords(axis_b_id)
        h_b_x, h_b_y = np.mean( h_b_coords[0::2] ),  np.mean( h_b_coords[1::2] )
        b = np.sqrt( (h_b_x - cx)**2 + (h_b_y - cy)**2 )

        angle = np.rad2deg( np.arctan2( cy - h_a_y, h_a_x - cx) )

        rings = self.canvas.itemcget(ringtext_id, 'text')

        return cx, cy, a, b, angle, rings, ell_coords


    def update_df(self):
        # iterate through all ellipses -- i.e. tokens with with 'main' tag
        tokenitems = self.canvas.find_withtag( 'main' )


    def on_handle_press(self, event):
        '''Begining drag of an handle'''
        # record the item and its location
        ids = self.canvas.find_closest(event.x, event.y)[0]
        tags = self.canvas.gettags( ids )
        self._drag_data["items"] = ids
        self._drag_data["x"] = event.x
        self._drag_data["y"] = event.y

    def on_handle_release(self, event):
        '''End drag of an handle'''
        # reset the drag information
        self._drag_data["items"] = None
        self._drag_data["x"] = 0
        self._drag_data["y"] = 0

    def on_handle_motion(self, event):
        '''Handle dragging of an handle'''
        # compute how much the mouse has moved
        oldx, oldy = self._drag_data["x"], self._drag_data["y"]
        delta_x = event.x - oldx
        delta_y = event.y - oldy
        # move the handle the appropriate amount
        self.canvas.move(self._drag_data["items"], delta_x, delta_y)

        # what are the tags for this particular handle
        tags = self.canvas.gettags( self._drag_data["items"] )
        tokentag = tags[0]
        cx, cy, a, b, angle, rings, coords  = self.retrieve_ellipse_info( tokentag )

        tokenitems = self.canvas.find_withtag( tokentag )
        [main_id, axis_a_id, axis_b_id, ringtext_id ]= tokenitems

        new_r = np.sqrt( (event.x -cx)**2 + (event.y - cy)**2 )
        new_angle = np.rad2deg( np.arctan2( cy-oldy, oldx-cx) )

        # which handle is currently being manipulated?
        if ("axis_a" in tags):
            b_coords = self.canvas.coords(axis_b_id)
            h_b_x, h_b_y = np.mean( b_coords[0::2] ),  np.mean( b_coords[1::2] )
            new_coords = poly_oval( cx, cy, new_r, b, angle=new_angle)
            h_b_x, h_b_y =  cx + b*np.sin(np.deg2rad(new_angle)),  cy + b*np.cos(np.deg2rad(new_angle))
            self.canvas.coords(axis_b_id, [ h_b_x-self.hr, h_b_y-self.hr, h_b_x+self.hr, h_b_y+self.hr] )
        elif ("axis_b" in tags):
            a_coords = self.canvas.coords(axis_a_id)
            h_a_x, h_a_y = np.mean( a_coords[0::2] ),  np.mean( a_coords[1::2] )
            new_angle = new_angle + 90   # a and b axes are offset by 90 degrees; angle is defined relative to a axis
            new_coords = poly_oval( cx, cy, a, new_r, angle=new_angle)
            h_a_x, h_a_y =  cx + a*np.cos(np.deg2rad(new_angle)),  cy - a*np.sin(np.deg2rad(new_angle))
            self.canvas.coords(axis_a_id, [ h_a_x-self.hr, h_a_y-self.hr, h_a_x+self.hr, h_a_y+self.hr] )
        else:
            print("Error: bad tags")
            assert(0==1)

        self.canvas.coords(main_id, new_coords)  # reassign coords for the entire ellipse (i.e. 'redraw')
        # record the new position
        self._drag_data["x"] = event.x
        self._drag_data["y"] = event.y


    def on_doubleclick(self, event):  # create a new ellipse
        cx, cy, a, b, angle, rings = event.x, event.y, 50, 50, 0, 1     # give it some default data
        self._create_token((cx, cy), (a, b), angle, rings, self.color)
        self.update_readout(None)

    def on_rightpress(self,event):
        obj_id = self.canvas.find_closest(event.x, event.y)[0]
        tags = self.canvas.gettags( obj_id )
        ringtext = self.canvas.itemcget(obj_id, 'text')
        result = askfloat("How many rings", "How many rings?", initialvalue=float(ringtext))

        self.canvas.itemconfigure(obj_id,text=str(result))
        self.canvas.focus_set()           # that dialog box stole the focus. get it back
        self.update_readout(None)

    def update_readout(self, event):
        mains = self.canvas.find_withtag( "main" )
        self.infostr = self.meta_file+'\n'+str(self.img_file)+'\n\n'

        new_df = pd.DataFrame(columns=self.df.columns)
        # first we update the dataframe info
        for main_id in mains:
            tokentag = self.canvas.gettags( main_id )[0]
            cx, cy, a, b, angle, rings, coords = self.retrieve_ellipse_info( tokentag )
            a,b,angle = fix_abangle(a,b,angle)
            [cx, cy, a, b, angle] = [int(round(x)) for x in [cx, cy, a, b, angle]]
            #x, cy, a, b, angle = int(round(cx,0)), int(round(cy,0)), int(round(a,0), round(b,0), round(angle,0)
            new_df = new_df.append({'cx':cx, 'cy':cy, 'a':a, 'b':b, 'angle':angle, 'rings':rings},ignore_index=True)

        self.df = new_df
        self.infostr += self.df.to_string(index=False,justify='left')      # then we output the dataframe info to a string
        self.canvas.itemconfigure(self.text, text=self.infostr)   # then we re-assign the text widget with the new string



@call_parse
def ellipse_editor(
    seq:Param("Ignore top-loss ordering and do sequential ordering", store_true),
    files:Param("Wildcard name for all CSV files to edit", str)='annotations/*.csv',
    imgbank:Param("Directory where all the (unlabeled) images are",str)='images/',
    tldir:Param("Directory where 'top losses' info is stored'",str)='top_losses/',
    ):
    global img_bank
    # typical command-line calling sequence:
    #  $ ./ellipse_editor.py *.csv
    files = ''.join(files)
    meta_file_list = sorted(glob.glob(files))
    img_bank = imgbank

    print("Instructions:")
    print(" Mouse bindings:")
    print("    - Double-click to create ellipse")
    print("    - Left-click and drag inside ellipse (but not on a number) to move ellipse")
    print("    - Left-click and drag 'handles' to resize/rotate ellipse (solid = 'a', hollow = 'b')")
    print("    - Right-click (or middle click) inside number to change ring count")
    print("    - Drag off-screen to destroy/delete ellipse")
    print(" Key bindings:")
    print("    - Right Arrow : Next file")
    print("    - Left Arrow  : Previous file")
    print("    - M           : Toggle display of predicted segmentation mask (if available)")
    print("    - R           : Toggle display of predicted ring counts (if available)")
    print("    - B           : Toggle display of predicted bounding boxes (if available)")
    print("    - S           : Save metadata")
    print("    - Q           : Quit")


    root = tk.Tk()
    root.title('espiownage: ellipse_editor')
    EllipseEditor(root, meta_file_list, img_bank=img_bank, tldir=tldir, seq=seq).pack(fill="both", expand=True)
    root.mainloop()




from fastcore.script import *
from pathlib import Path
import glob
import os
import pandas as pd
import numpy as np
import cv2
from PIL import Image
import json
from espiownage.core import *
#from functools import partial
#import multiprocessing as mp

""" Generates bounding box info from our edited CSV files.  It outputs in two formats:
    1. .json: COCO(-TINY)-style bbox JSON file
    2. .csv: All the annotations in one big long CSV: see this notebook where we use this format:
        https://colab.research.google.com/drive/1bi8PYLFexoEcNKRClul0X3U6JRLKaH7M?usp=sharing
"""

def gen_coco_json(meta_file_list, bboxdir, step, reg, maxrings=11, allone=True):
    """ sample file format is a dict like...
    sample_coco_dict = {
        "categories": [{"id": 62, "name": "chair"}, {"id": 63, "name": "couch"}, {"id": 72, "name": "tv"}, {"id": 75, "name": "remote"}, {"id": 84, "name": "book"}, {"id": 86, "name": "vase"}],
        "images": [{"id": 542959, "file_name": "000000542959.jpg"}, {"id": 129739, "file_name": "000000129739.jpg"}],
        "annotations": [{"image_id": 542959, "bbox": [32.52, 86.34, 8.53, 9.41], "category_id": 62}, {"image_id": 542959, "bbox": [98.12, 110.52, 1.95, 4.07], "category_id": 86}, {"image_id": 542959, "bbox": [91.28, 51.62, 3.95, 5.72], "category_id": 86}, {"image_id": 542959, "bbox": [110.48, 110.82, 14.55, 15.22], "category_id": 62}, {"image_id": 542959, "bbox": [96.63, 50.18, 18.67, 13.46], "category_id": 62}, {"image_id": 542959, "bbox": [0.69, 111.73, 11.8, 13.06], "category_id": 62}]
    }"""
    json_filename = bboxdir+'/coco_bboxes.json'
    print(f"Generating COCO-style JSON file {json_filename} ...")
    coco_dict = {}

    # categories
    if allone:
        print("allone=True: Treating all objects as same class")
        coco_dict["categories"] = [{"id": 0, "name": "AN"}]
    elif reg:
        print('Regression model: 1 class, called "rings"')
        coco_dict["categories"] = [{"id": 0, "name": "rings"}] # probably won't work
    else:
        coco_dict["categories"] = [{"id": x, "name":str((x+1)*step)} for x in range(int(round(maxrings/step)))]

    # images
    coco_dict["images"] = [{"id":i, "file_name":os.path.basename(str(meta_to_img_path(x)))} for i,x in enumerate(meta_file_list)]

    # annotations
    ann_list = []
    for i, meta_file in enumerate(meta_file_list):
        this_df = meta_to_df(meta_file)
        image = os.path.basename(str(meta_to_img_path(meta_file)))
        #print("meta_file = ",meta_file)
        for index, row in this_df.iterrows():
            [cx, cy, a, b, angle] = [x for x in [row['cx'], row['cy'], row['a'], row['b'], row['angle']]]
            bbox = ellipse_to_bbox(cx, cy, a, b, angle, coco=True)
            rings = round(float(row['rings']),2)
            assert rings <= maxrings
            if (rings > 0) and (bbox is not None):
                if allone:
                    category_id = 0
                elif reg:
                    category_id = rings
                else:
                    int(round(maxrings/step))
                this_ann = {"image_id":image, "bbox": bbox, "category_id":category_id}
                ann_list = ann_list + [this_ann]
    coco_dict["annotations"] = ann_list

    #print("coco_dict =\n",coco_dict)
    with open(json_filename, 'w') as fp:
        json.dump(coco_dict, fp)
    return


def gen_long_csv(
    files_str,          # original string specifying meta / CSV files
    meta_file_list,     # list of those files, with paths
    bboxdir,            # where we're writing to
    step,               # quantize ring counts via this bin size
    reg,                # regression model?
    obpr=False,         # "one box per ring" mode. every thing is an object "ring"
    allone=True,        # all antinodes get marked as the same class: "AN" for antinode
    maxrings=11,        # used for quantization.
    quiet=True,        # don't list every name created
    ):
    out_csv_filename = bboxdir+'/annotations_obpr.csv' if obpr else bboxdir+'/annotations.csv'
    if allone: print("allone=True: Treating all objects as same class")
    print(f"Generating long CSV {out_csv_filename} ...")

    width, height = 512, 384  # image dims
    final_col_names = ['filename','width', 'height', 'label', 'xmin', 'ymin', 'xmax', 'ymax']

    # Goal: make one big long list, turn it into a DataFrame, and then write it
    ann_list = []
    for i, meta_file in enumerate(meta_file_list):
        if (not quiet): print("meta_file = ",meta_file, ", quiet = ",quiet)
        this_df = meta_to_df(meta_file)
        image_file = os.path.basename(str(meta_to_img_path(meta_file)))
        this_df['filename'] = image_file

        for index, row in this_df.iterrows(): #convert to bboxes
            [cx, cy, a, b, angle] = [x for x in [row['cx'], row['cy'], row['a'], row['b'], row['angle']]]
            rings = round(float(row['rings']),2)
            assert rings <= maxrings
            if rings > 0:
                if not obpr:
                    bbox = ellipse_to_bbox(cx, cy, a, b, angle, coco=False)
                    if bbox is not None:
                        if allone: label = 'AN'
                        elif reg: label = rings
                        else: label = ring_float_to_class_int(rings, step=step)
                        line_list = [image_file, width, height, label, bbox[0], bbox[1], bbox[2], bbox[3]]
                        ann_list.append(line_list)
                else:                           # one box per ring (rounded as integers)
                    rings_int, label = round(int(rings)), 'ring'
                    line_list = []
                    for i in range(rings_int,0,-1):  # counts down to 1, 0 is not included per Python norms
                        _a, _b = (i/rings_int)*a, (i/rings_int)*b
                        bbox = ellipse_to_bbox(cx, cy, _a, _b, angle, coco=False)
                        if bbox is not None:
                            line_list = ([image_file, width, height, label, bbox[0], bbox[1], bbox[2], bbox[3]])
                            ann_list.append(line_list)



    print("   Creating data frame")
    new_df = pd.DataFrame(ann_list, columns=final_col_names)
    new_df = new_df[final_col_names]  # just to force ordering
    new_df.to_csv(out_csv_filename, index=False)
    return


@call_parse
def gen_bboxes(
    reg:Param("Set this for regression model (1 class, no steps)", store_true),
    notallone:Param("All objects DON'T get assigned to same class: 'AN' for antinode", store_true),
    obpr:Param("Set this for one box per ring", store_true),
    notquiet:Param("Don't list every filename created", store_true),
    files:Param("Wildcard name for all (ellipse) CSV files to read", str)='annotations/*.csv',
    bboxdir:Param("Directory to write bboxes to",str)='bboxes',
    step:Param("For classification model: Step size / resolution / precision of ring count",float)=1,
    ):

    mkdir_if_needed(bboxdir)

    files = ''.join(files)  # convert to str
    meta_file_list = sorted(glob.glob(files)) # list of all annotation .csv files for ellipses


    gen_long_csv(files, meta_file_list, bboxdir, step, reg, obpr=obpr, allone=(not notallone), quiet=(not notquiet))
    gen_coco_json(meta_file_list, bboxdir, step, reg, allone=(not notallone))

    return



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = ['sysinfo', 'mkdir_if_needed', 'get_data', 'get_checkpoint', 'meta_to_img_path', 'meta_to_mask_path',
           'meta_from_str', 'combine_file_and_tl_lists', 'meta_to_df', 'fix_abangle', 'draw_ellipse', 'ellipse_to_bbox',
           'ring_float_to_class_int', 'crop_to_bbox', 'is_in_box', 'acc_reg', 'acc_reg05', 'acc_reg07', 'acc_reg1',
           'acc_reg15', 'acc_reg2', 'kfold_split']

# Cell
import cv2
from PIL import Image
import numpy as np
from pathlib import Path
import os
import pandas as pd
from fastai.torch_core import flatten_check
import re
import math
import torch

# Cell
#slow
def sysinfo():
    "Prints out versions of PyTorch, CUDA, name of GPU and host machine (for logging)"
    try:
        tv, cv = torch.__version__, torch.version.cuda
        tv = re.sub('\+cu.*','',tv)
        TORCH_VERSION = 'torch'+tv[0:-1]+'0'
        CUDA_VERSION = 'cu'+cv.replace('.','')
        print(f"TORCH_VERSION={TORCH_VERSION}; CUDA_VERSION={CUDA_VERSION}")
        print(f"CUDA available = {torch.cuda.is_available()}, Device count = {torch.cuda.device_count()}, Current device = {torch.cuda.current_device()}")
        print(f"Device name = {torch.cuda.get_device_name()}")
    except:
        print("No CUDA info available!  (That's bad)")
    print("hostname:",os.uname()[1])

# Cell
def mkdir_if_needed(path:str):
    try:                # go ahead and try to make the the directory
        os.makedirs(path)
    except FileExistsError: pass
    except OSError as exception:
        if exception.errno != errno.EEXIST:  # ignore error if dir already exists
            raise

# Cell
def get_data(dataset_name='cleaner',    # 'cleaner','preclean','spnet','cyclegan','fake'
             force_download=False,      # re-downloads even if you've already got one
             dest_root='~/.espiownage/data', # where the data gets saved
            ):
    "We 'anonymized' our URLs for double-blind review. But this screwed up fastai's downloader so we wrote one"
    assert dataset_name in ['cleaner','preclean','spnet','cyclegan','fake']
    urls = {}
    urls['cleaner']  = 'https://www.dropbox.com/sh/24eqya4frqg5i7c/AAAVEsRhmc06J_rnFUcId8-ga/data/espiownage-cleaner.tgz'
    urls['preclean'] = 'https://www.dropbox.com/sh/24eqya4frqg5i7c/AADxzcSWr0-5WlZwqM81KNzVa/data/espiownage-preclean.tgz'
    urls['spnet']    = 'https://www.dropbox.com/sh/24eqya4frqg5i7c/AACiVHrATAo6laLsRWvO2dvxa/data/espiownage-spnet.tgz'
    urls['cyclegan'] = 'https://www.dropbox.com/sh/24eqya4frqg5i7c/AABf0H1J4ama2s3WGGFHv3MFa/data/espiownage-cyclegan.tgz'
    urls['fake']     = 'https://www.dropbox.com/sh/24eqya4frqg5i7c/AAAQK6S36pMWkxd-XrXqcf0ma/data/espiownage-fake.tgz'
    data_dir = f'espiownage-{dataset_name}'
    dest_root = os.path.expanduser(dest_root)+'/'
    mkdir_if_needed(dest_root)
    dest = dest_root+data_dir
    if (not os.path.exists(dest)) or (force_download):
        cmd = f"curl -L -o {dest+'.tgz'} {urls[dataset_name]}; tar xfz {dest+'.tgz'} -C {dest_root}"
        print(cmd)
        os.system(cmd)
    return Path(dest_root+data_dir)

# Cell
def get_checkpoint(
        model_type='segreg', # 'segreg', 'rings', 'bboxes' . trained on real data only
        learner_or_model=None, # for segreg & rings you need to pass in a defined fastai Learner; for bboxes we grab a whole model
        force_download=False,  # re-downloads even if you've already got one
        dest_root='~/.espiownage/models', # where the model weights gets saved
        ):
    "For inference: Grab checkpoint of a pretrained espiownage model"
    assert model_type in ['segreg', 'rings', 'bboxes']
    urls = {}
    urls['segreg']  = ''
    urls['rings'] = ''
    urls['bboxes']    = ''
    if 'bboxes' == model_type:
        print("Warning: the IceVision install is incompatible with the fastai version used for the seg-reg and ring counting")
        print("For bounding boxes you (probably) need a dedicated Python environment for that.")
    else:
        assert learner_or_model is not None, "You need to define the Learner so we can load the weights"

    print("FAILURE: This isn't ready yet. Returning")
    return None, model

    data_dir = f'espiownage-{dataset_name}'
    dest_root = os.path.expanduser(dest_root)+'/'
    mkdir_if_needed(dest_root)
    dest = dest_root+data_dir
    if (not os.path.exists(dest)) or (force_download):
        cmd = f"curl -L -o {dest+'.tgz'} {urls[dataset_name]}; tar xfz {dest+'.tgz'} -C {dest_root}"
        print(cmd)
        os.system(cmd)
    return Path(dest_root+data_dir)

# Cell
def meta_to_img_path(
    meta_file:str, # filename of .csv file with annotations
    img_bank='images/',         # alternate location of image files, if not in same dir as meta file
    ):
    '''Suggest the image file that corresponds with an annotation CSV file'''
    meta_file, img_dir = Path(meta_file), Path(os.path.expanduser(img_bank))
    img_file = meta_file.with_suffix('.png')  # check same directory as meta first
    if os.path.exists(img_file): return img_file
    return img_dir / Path(img_file.name)  # return from image storage directory

# Cell
def meta_to_mask_path(
    meta_file:str, # filename of .csv file with annotations
    mask_dir='masks/',  # output directory; assumed to exist
    ):
    "provides name for segmentation mask file"
    csv_path = Path(meta_file)
    return Path(mask_dir + csv_path.stem+'_P.png')  # _P because that's what CAMVID dataset does

# Cell
def meta_from_str(s):
    "e.g., 06241902_proc_00510_0_45_88_236_1.0 -> 06241902_proc_00510.csv"
    # grab numbers on either side of _proc_
    s = s.split('.')[0]
    splits = s.split('_')
    return splits[0]+'_'+splits[1]+'_'+splits[2]+'.csv'

# Cell
def combine_file_and_tl_lists(file_list, top_loss_list):
    """for ellipse editor, but adding here:
    Go through tl list, and if elem e is not in file list, don't copy
    then go through file_list, and if e is already in, don't add"""
    if [] == top_loss_list: return file_list
    new_list = []
    basenames = [os.path.basename(f) for f in file_list]
    fldirname = os.path.dirname(file_list[0])
    meta_top_loss_list = [meta_from_str(x) for x in top_loss_list] # convert from whatever janky name is in tl list
    for e in meta_top_loss_list:
        if e in basenames: new_list.append(fldirname+'/'+e)
    for e in file_list:
        if e not in new_list: new_list.append(e)
    return new_list

# Cell
def meta_to_df(
    meta_file,  # csv file of ellipse data for an image
    ):
    "Reads in an espiownage/SPNet CSV file of ellipse data and returns a Pandas DataFrame"
    col_names = ['cx', 'cy', 'a', 'b', 'angle', 'rings']
    df = pd.read_csv(meta_file, header=None, names=col_names)
    df.drop_duplicates(inplace=True)  # sometimes the data from Zooniverse has duplicate rows
    for index, row in df.iterrows() : # fixup stuff
        [cx, cy, a, b, angle] = [int(round(x)) for x in [row['cx'], row['cy'], row['a'], row['b'], row['angle']]]
        a, b, angle = fix_abangle(a,b,angle)
        row['cx'], row['cy'], row['a'], row['b'], row['angle'] = cx, cy, a, b, angle
    return df

# Cell
def fix_abangle(
    a:float, # semimajor axis
    b:float, # semiminor axis
    angle:float,  # orientation angle in degrees
    ):
    "Makes sure semimajor axis > semiminor axis, and angles are consistent"
    if b > a:
        a, b, angle = b, a, angle+90
    if angle < 0: angle += 180
    elif angle >= 180: angle -= 180
    return a, b, angle

# Cell
def draw_ellipse(
    img,         # a cv2 image, *not* a PIL image (similar for grayscale but not RGB)
    center:tuple,      # (cx, cy) tuple
    axes:tuple,        # (a,b) semimajor & minor axes
    angle:float,       # orientation angle in degrees
    color=(0),   # color to draw. tuple or int
    thickness=2, # thickness ofthe lines we draw
    filled=False,  # whether to draw the ellipse as filled or not
    lineType=cv2.LINE_8,  # as opposed to LINE_AA, typically we DON'T want antialiasing for this app
    startAngle=0, endAngle=360, # arc-angles. should stay at 0 & 360 for full ellipses.
    shift=0, #10,      # shift is for sub-pixel resolution and AA figures. don't need it.
    ):
    """"Draws an ellipse into image.    """
    center = [int(round(x* 2**shift)) for x in center]
    axes = [int(round(x* 2**shift)) for x in axes]
    if filled: lineType, thickness = cv2.FILLED, -1
    ellipse = cv2.ellipse(
        img, center, axes, -angle,   # -angle because the web interface is "upside down"
        startAngle, endAngle, color,
        thickness, lineType, shift)
    return ellipse

# Cell
def ellipse_to_bbox(
    cx:float, # x-coordinate of center of ellipse
    cy:float, # y-coordinate of center of ellipse
    a:float,  # semimajor axis
    b:float,  # semiminor axis
    angle_deg:float,  # orientation angle in degrees
    coco=False, # COCO style bbox has last 2 nums as width & height of bbox
    width=512, height=384, # image dimensions for clipping
    clip=True, # clip values at max values of image width & height
    nozero=True,  # Lots of downstream apps hate zero-dimension bounding box. This returns None
    ):
    "converts ellipse to bounding box"
    rad = np.radians(angle_deg)
    a2, b2, cos2, sin2 = [x**2 for x in [a, b, np.cos(rad), np.sin(rad)]]
    delta_x, delta_y = np.sqrt(a2*cos2 + b2*sin2), np.sqrt(a2*sin2 + b2*cos2)
    xmin, xmax = cx - delta_x,  cx + delta_x
    ymin, ymax = cy - delta_y,  cy + delta_y
    if (xmin > xmax): xmin, xmax = xmax, xmin  # error correction, swap
    if (ymin > ymax): ymin, ymax = ymax, ymin  # swap
    if clip:
        xmin, xmax = np.clip(xmin, 0, width),  np.clip(xmax, 0, width)
        ymin, ymax = np.clip(ymin, 0, height), np.clip(ymax, 0, height)
    if coco: return [round(x, 2) for x in [xmin, ymin, xmax-xmin, ymax-ymin]] # coco is a list, floats are ok
    bbox = int(xmin), int(ymin), int(xmax), int(ymax)
    # Image.crop does not like zero-size dimensions but its error message is cryptic
    if ((xmax-xmin > 0) and (ymax-ymin > 0)) or (not nozero): return bbox
    else:
        print(f"ellipse_to_bbox: Error: zero-dim bbox = {bbox}. Returning None.")
        return None

# Cell
def ring_float_to_class_int(rings:float, step=0.1):
    """Ring value rounded to classifier value; rounded to nearest step size"""
    return round(rings/step)

# Cell
def crop_to_bbox(
    img,        # an image (PIL preferred, but will convert from cv2 if neede)
    bbox,       # [xmin, ymin, d3, d4] where d3,d4 are either xmax,ymax (default) or see coco (below)
    coco=False, # COCO style input bbox has last 2 nums as width & height of bbox
    width=512, height=384, # image dimensions for clipping
    clip=True,  # clip values at max values of image width & height
    ):
    "Crops an image to bbox, returns cropped image"
    if isinstance(img,np.ndarray): img = Image.fromarray(img)  # convert cv2 image to PIL
    xmin, ymin = bbox[0], bbox[1]
    if coco: xmax, ymax = bbox[0]+bbox[2], bbox[1]+bbox[3]
    else:    xmax, ymax = bbox[2], bbox[3]
    if clip:
        xmin, xmax = np.clip(xmin, 0, width),  np.clip(xmax, 0, width)
        ymin, ymax = np.clip(ymin, 0, height), np.clip(ymax, 0, height)
    # check ordering
    if (xmin > xmax): xmin, xmax = xmax, xmin  # error correction, swap
    if (ymin > ymax): ymin, ymax = ymax, ymin  # swap
    crop_bb = (int(xmin), int(ymin), int(xmax), int(ymax))

    # Image.crop does not like zero-size dimensions but its error message is cryptic
    if (xmax-xmin > 0) and (ymax-ymin > 0): return img.crop(crop_bb)
    else:
        print(f"crop_to_bbox: Error: zero-dim crop request, crop_bb = {crop_bb}. Returning None.")
        return None

# Cell
def is_in_box(
    p,     # a point as a (x,y) coordinate pair
    bb,    # a bounding box of the form [xmin,ymin,xmax,ymax]
    ):
    "Is this point within that bounding box?"
    x, y = p
    return (x>=bb[0]) and (x<=bb[2]) and (y>=bb[1]) and (y<=bb[3])

# Cell
def acc_reg(inp, targ, bin_size=1):
    "Accuracy for regression: Are we within +/- bin_size?"
    inp,targ = flatten_check(inp,targ) # https://docs.fast.ai/metrics.html#flatten_check
    where_correct = (inp-targ).abs() < bin_size
    return where_correct.float().mean()

# Cell
def acc_reg05(inp, targ): return acc_reg(inp, targ, bin_size=0.5)

# Cell
def acc_reg07(inp, targ): return acc_reg(inp, targ, bin_size=0.7)

# Cell
def acc_reg1(inp, targ): return acc_reg(inp, targ, bin_size=1)

# Cell
def acc_reg15(inp, targ): return acc_reg(inp, targ, bin_size=1.5)

# Cell
def acc_reg2(inp, targ): return acc_reg(inp, targ, bin_size=2)

# Cell
def kfold_split(
    data,         # list of indices or data points to split into train & val
    k,            # where we are in the iteration
    nk = 5         # number of folds
    ):
    "Breaks one list into train & val as per k-fold cross-validation. For use with fastai IndexSplitter (throw out train_list) or IceVision's FixedSplitter"
    n = len(data)
    val_size = int(round(n/nk))
    if k < nk-1:
        val_list = data[k*val_size:(k+1)*val_size]
        train_list = data[0:k*val_size] + data[(k+1)*val_size:n]
    else:   # last one might be a bit different
        val_list = data[k*val_size:]
        train_list = data[0:-len(val_list)]
    return train_list, val_list



from fastcore.script import * 
from pathlib import Path
import glob
import os, shutil
import pandas as pd
import numpy as np
import cv2
from PIL import Image
from espiownage.core import *
from functools import partial
import multiprocessing as mp
import sys

""" Note that (normally) we don't care about the actual images. we're just generating
    image masks from the annotations
"""


imgbank, ann_img_dir = "images/", "annotated_images/"
all_colors = {0}

def handle_one_file(meta_file_list, # list of all the csv files
    mask_dir,                # output directory, where to write mask files to
    step,                   # resolution of (float) rings, when converting to int
    #imglinks,               # boolean on whether or not to create links to original images
    allone,                 # set all values to one for non-background
    cp_ann_imgs,            # hack to make directory of only images for which annotations exist
    quiet,                   # don't print many status messages
    i,                      # index of which meta file we'll read from
    height=512, width=384):  # image dimensions
    global all_colors, ann_img_dir

    meta_file = meta_file_list[i]
    col_names = ['cx', 'cy', 'a', 'b', 'angle', 'rings']
    df = pd.read_csv(meta_file, header=None, names=col_names)

    # progress message
    mask_path = meta_to_mask_path(meta_file, mask_dir=mask_dir+'/')
    if not quiet:
        print(f"{i}/{len(meta_file_list)}: meta_file = {meta_file}, mask_path = {mask_path} ",flush=True)

    df.drop_duplicates(inplace=True)  # sometimes the data from Zooniverse has duplicate rows
    img = np.zeros((width, height), dtype=np.uint8)  # blank black image, dimensions are wonky
    for index, row in df.iterrows() :
        [cx, cy, a, b, angle] = [int(round(x)) for x in [row['cx'], row['cy'], row['a'], row['b'], row['angle']]]
        rings = round(float(row['rings']),2)
        a, b, angle = fix_abangle(a,b,angle)
        if (rings > 0):
            if rings > 11: # saw an error once
                print(f"Sever warning for file {meta_file}: rings = {rings}.  Aborting")
                sys.exit(1)
            color = 1 if allone else ring_float_to_class_int(rings, step=step)
            all_colors = all_colors.union({color})
            img = draw_ellipse(img, (cx,cy), (a,b), angle, color=color, filled=True)

    #all_colors = all_colors.union(set(np.array(img).flatten()))  # super sanity check but slow
    # cv2.imwrite(str(mask_path), img)  Don't write as cv2, write as PIL
    pil_image = Image.fromarray(img)
    pil_image = pil_image.save(str(mask_path))


    if cp_ann_imgs:
        src_img = meta_to_img_path(meta_file)
        dest_img = ann_img_dir+os.path.basename(src_img)
        shutil.copyfile(src_img, dest_img)

    return img



@call_parse
def gen_masks(
    allone:Param("All objects get assigned to class 1", store_true),
    quiet:Param("Suppress output log", store_true),
    cp_ann_imgs:Param("make directory of only images for which annotations exist (to annotated_images/)", store_true),
    files:Param("Wildcard name for all CSV files to edit", str)='annotations/*.csv',
    maskdir:Param("Directory to write segmentation masks to",str)='masks/',
    step:Param("Step size / resolution / precision of ring count",float)=1,
    ):
    "Generate segmentation masks for all annotations"
    global all_colors, ann_img_dir

    mkdir_if_needed(maskdir)
    if cp_ann_imgs: mkdir_if_needed(ann_img_dir)

    files = ''.join(files)
    meta_file_list = sorted(glob.glob(files))

    parallel = True  # could leave off. it's not that slow, really
    if not parallel:
        for i in range(len(meta_file_list)):
            handle_one_file(meta_file_list, maskdir, step, allone, cp_ann_imgs, quiet, i)
        print("all_colors = ",sorted(list(all_colors))) # Very handy
    else:
        # parallel processing
        wrapper = partial(handle_one_file, meta_file_list, maskdir, step, allone, cp_ann_imgs, quiet)
        pool = mp.Pool(mp.cpu_count())
        results = pool.map(wrapper, range(len(meta_file_list)))
        pool.close()
        pool.join()

    return



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"sysinfo": "00_core.ipynb",
         "mkdir_if_needed": "00_core.ipynb",
         "get_data": "00_core.ipynb",
         "get_checkpoint": "00_core.ipynb",
         "meta_to_img_path": "00_core.ipynb",
         "meta_to_mask_path": "00_core.ipynb",
         "meta_from_str": "00_core.ipynb",
         "combine_file_and_tl_lists": "00_core.ipynb",
         "meta_to_df": "00_core.ipynb",
         "fix_abangle": "00_core.ipynb",
         "draw_ellipse": "00_core.ipynb",
         "ellipse_to_bbox": "00_core.ipynb",
         "ring_float_to_class_int": "00_core.ipynb",
         "crop_to_bbox": "00_core.ipynb",
         "is_in_box": "00_core.ipynb",
         "acc_reg": "00_core.ipynb",
         "acc_reg05": "00_core.ipynb",
         "acc_reg07": "00_core.ipynb",
         "acc_reg1": "00_core.ipynb",
         "acc_reg15": "00_core.ipynb",
         "acc_reg2": "00_core.ipynb",
         "kfold_split": "00_core.ipynb"}

modules = ["core.py",
           "scripts.py"]

doc_url = "https://drscotthawley.github.io/espiownage/"

git_url = "https://github.com/drscotthawley/espiownage/tree/master/"

def custom_doc_links(name): return None




from fastcore.script import *
from pathlib import Path
import glob
import os
import pandas as pd
import numpy as np
import cv2
from PIL import Image
from espiownage.core import *
from functools import partial
import multiprocessing as mp
import sys

""" Generates cropped images of individual antinodes using annotations
"""

def handle_one_file(meta_file_list, # list of all the csv files
    outdir,                # output directory, where to write mask files to
    i,                      # index of which meta file we'll read from
    height=512, width=384):  # image dimensions

    meta_file = meta_file_list[i]
    print(f"{i}/{len(meta_file_list)}: meta_file = {meta_file}",flush=True)

    img_path = meta_to_img_path(meta_file)
    img = Image.open(img_path)

    # read meta csv file for all rings
    col_names = ['cx', 'cy', 'a', 'b', 'angle', 'rings']
    df = pd.read_csv(meta_file, header=None, names=col_names)
    df.drop_duplicates(inplace=True)  # sometimes the data from Zooniverse has duplicate rows
    for index, row in df.iterrows() :
        [cx, cy, a, b, angle] = [int(round(x)) for x in [row['cx'], row['cy'], row['a'], row['b'], row['angle']]]
        rings = round(float(row['rings']),2)
        a, b, angle = fix_abangle(a,b,angle)
        if (rings > 0):
            bb = ellipse_to_bbox(cx, cy, a, b, angle)
            if bb is not None:
                img_cropped = crop_to_bbox(img, bb)
                if img_cropped is not None:
                    out_file = outdir+'/'+str(Path(meta_file).stem)+f"_{bb[0]}_{bb[1]}_{bb[2]}_{bb[3]}_{rings}.png"
                    img_cropped.save(out_file)
    return



@call_parse
def gen_crops(
    files:Param("Wildcard name for all CSV files to edit", str)='annotations/*.csv',
    outdir:Param("Directory to write output cropped images to",str)='crops/',
    ):
    "Generate cropped images for all annotations"

    mkdir_if_needed(outdir)

    files = ''.join(files)
    meta_file_list = sorted(glob.glob(files))

    parallel = True  # not too slow sequential but parallel=True is good
    if not parallel:
        for i in range(len(meta_file_list)):
            handle_one_file(meta_file_list, outdir, i)
    else:
        # parallel processing
        wrapper = partial(handle_one_file, meta_file_list, outdir)
        pool = mp.Pool(mp.cpu_count())
        results = pool.map(wrapper, range(len(meta_file_list)))
        pool.close()
        pool.join()

    return




# Generates 'fake' images akin to the ESPI images of steelpan drums, from
# https://www.zooniverse.org/projects/achmorrison/steelpan-vibrations
#

# num_rings can be between 0 and 11.  0 means there is no antinode there
# You can also have num_antinodes=0, but this produces more of a 'skewed' dataset
#   than if you use num_rings=0 instead and let num_antinodes=1

# note that this doesn't require any GPU usage, just a lot of disk usage

# Added multiprocesing: runs lots of processes to cut execution time down

from fastcore.script import *

import numpy as np
import cv2
import random
import os
import time
from functools import partial
import multiprocessing as mp
from espiownage.core import *

import sys, traceback
from shutil import get_terminal_size
import glob

winName = 'ImgWindowName'
imWidth = 512
imHeight = 384

meta_extension = ".csv"     # file extension for metadata files

# Define some colors: openCV uses BGR instead of RGB
blue = (255,0,0)
red = (0,0,255)
green = (0,255,0)
white = (255)
black = (0)
grey = (128)

blur_prob = 0.3    # probability that an image gets blurred

min_line_width = 4  # number of pixels per each ring (dark-light pair)

pad = "       "

def bandpass_mixup(img_fake, path_real=os.path.expanduser('~/datasets/espiownage-data/images/')):
    '''
    For more realistic-looking images (still not as good as StyleGAN),
    replace low & high frequency components ('background')
    of fake images using those components from real images
    '''

    # get a random background from the group of 'true' images
    file_true = random.choice(glob.glob(path_real+'/*.png'))
    img_true = cv2.imread(file_true, cv2.IMREAD_GRAYSCALE)
    # maybe flip the image
    flipchoice = np.random.choice([-1,0,1,2])
    if (flipchoice != 2):
        img_true = cv2.flip(img_true, flipchoice)

    # take fourier transforms of fake and true images
    dft_true = cv2.dft(np.float32(img_true),flags = cv2.DFT_COMPLEX_OUTPUT)
    dft_shift_true = np.fft.fftshift(dft_true)    # center the "dc" part of image

    dft_fake = cv2.dft(np.float32(img_fake),flags = cv2.DFT_COMPLEX_OUTPUT)
    dft_shift_fake = np.fft.fftshift(dft_fake)    # center the "dc" part of image

    # Set up a filter: Keep the Lows and the Highs
    # create a rectagular mask first, center square is 1, remaining all zeros. LPF
    rows, cols = img_fake.shape
    crow,ccol = rows//2 , cols//2
    wl, wh = 8, 0     #  width for LPF and HPF respectively
    mask = np.zeros((rows,cols,2),np.uint8)
    mask[crow-wl:crow+wl, ccol-wl:ccol+wl] = 1   # LPF
    if wh > 0:
        mask[0:wh,:] = 1    # HPF
        mask[-wh:,:] = 1    # HPF
        mask[:,0:wh] = 1    # HPF
        mask[:,-wh:] = 1    # HPF
    fshift = np.random.rand()*3*dft_shift_true*mask + (1-mask)*dft_shift_fake   # L/H from true, mids from fake

    # inverse DFT
    f_ishift = np.fft.ifftshift(fshift)
    img_back = cv2.idft(f_ishift)
    img_back = cv2.magnitude(img_back[:,:,0],img_back[:,:,1])
    cv2.normalize(img_back, img_back, 0, 255, cv2.NORM_MINMAX)

    return np.clip(img_back, 0, 255)


def blur_image(img, kernel_size=7):
    if (0==kernel_size):
        return img
    new_img = img#.copy()
    new_img = cv2.GaussianBlur(img,(kernel_size,kernel_size),0)
    return new_img


def draw_waves(img):  # TODO make intensity vary smoothly
    xs = np.arange(0, imWidth)
    ys = np.arange(0, imHeight)

    amp = random.randint(10,200)
    x_wavelength = random.randint(100,int(imWidth/2))
    thickness = random.randint(15,40)
    slope = 3*(np.random.rand()-.5)
    y_spacing = random.randint(thickness + thickness*int(np.abs(1.5*slope)), int(imHeight/3))
    numlines = 60+int(imHeight/y_spacing)

    for j in range(numlines):   # skips y_spacing in between drawings
        y_start = j*y_spacing - img.shape[1]*abs(slope)
        pts = []
        for i in range(len(xs)):
            pt = [ int(xs[i]), int(y_start + slope*xs[i]+ amp * np.cos(xs[i]/x_wavelength))]
            pts.append(pt)
        pts = np.array(pts, np.int32)
        cv2.polylines(img, [pts], False, black, thickness=thickness)
    return


# wrapper for legacy code with different calling sequence than newer code
def get_ellipse_box(center, axes, angle):  # converts ellipse to bounding box
    return ellipse_to_bbox( center[0], center[1], axes[0], axes[1], angle)



def draw_rings(img,center,axes,angle=45,num_rings=5.5):  # TODO make intensity vary smoothly
    num_drawrings, thickness = max(axes), 1
    if num_rings < 0.2: num_rings = 0.2+0.2*np.random.rand()
    phase = 2*np.pi*np.random.rand()
    minc, maxc = random.randint(0,60), random.randint(150,250)
    for j in range(num_drawrings):
        color = int(minc + (maxc-minc)*np.sin(2*np.pi*num_rings * j/num_drawrings + phase))
        thisring_axes = [axes[i] * (j+1)*1.0/(num_drawrings+1) for i in range(len(axes))]
        ellipse = draw_ellipse(img,center,thisring_axes,angle,color=color, thickness=thickness)
    return ellipse   # returns outermost ellipse




def does_overlap( a, b):
    if (a[2] < b[0]): return False # a is left of b
    if (a[0] > b[2]): return False # a is right of b
    if (a[3] < b[1]): return False # a is above b
    if (a[1] > b[3]): return False # a is below b
    return True

def does_overlap_previous(box, boxes_arr):
    # returns true if bounding box of new ellipse (ignoring angle) would
    # overlap with previous ellipses
    if ([] == boxes_arr):
        return False
    for i in range(len(boxes_arr)):
        #print("                boxes_arr[",i,"] = ",boxes_arr[i])
        if (does_overlap( box, boxes_arr[i])):
            return True
    return False



def draw_antinodes(img,num_antinodes=1):
    boxes_arr = []
    caption = ""

    if (num_antinodes==0):
        caption = "{0},{1},{2},{3},{4},{5}".format( 0,  0,    0,  0,    0,   0.0)  # as per @achmorrison's format

    for an in range(num_antinodes): # draw a bunch of antinodes

        axes = (random.randint(15,int(imWidth/3.5)), random.randint(15,int(imHeight/3.5)))   # semimajor and semiminor axes of ellipse
        axes = sorted(axes, reverse=True)   # do descending order, for definiteness. i.e. so a > b


        # TODO: based on viewing real images: for small antinodes, number of rings should also be small
        max_rings = min(axes[1] // 8, 11)                  # '8' chosen from experience looking at the data
        num_rings = np.random.uniform(low=0.5,high=11.0)            # well say that an antinode has at least 1 ring

        # make sure line width isn't too small to be resolved
        if (axes[1]/num_rings < min_line_width):
            num_rings = axes[1] / min_line_width

        center = (random.randint(axes[0], imWidth-axes[0]),
            random.randint(axes[1], imHeight-axes[1]))
        angle = random.randint(1, 179)       # ellipses are symmetric after 180 degree rotation
        box = get_ellipse_box(center, axes, angle)

        # make sure they don't overlap, and are in bounds of image
        # TODO: the following random placement is painfully inefficient
        trycount, maxtries = 0, 2000
        while (   ( (True == does_overlap_previous(box, boxes_arr))
            or (box[0]<0) or (box[2] > imWidth)
            or (box[1]<0) or (box[3] > imHeight)  ) and (trycount < maxtries) ):
            trycount += 1
            # if there's a problem, then generate new values - "Re-do"
            axes = (random.randint(25, int(imWidth/3)), random.randint(25, int(imHeight/3)))
            axes = sorted(axes, reverse=True)   # do descending order
            # make sure line width isn't too small to be resolved
            if (axes[1]/num_rings < min_line_width):
                num_rings = axes[1] / min_line_width

            center = (random.randint(axes[0], imWidth-axes[0]),
                random.randint(axes[1], imHeight-axes[1]))
            angle = random.randint(1, 180)
            box = get_ellipse_box(center, axes, angle)

        success = False
        if (trycount < maxtries):
            draw_rings(img, center, axes, angle=angle, num_rings=num_rings)
            this_caption = "{0},{1},{2},{3},{4},{5}".format(center[0], center[1],axes[0], axes[1], angle, round(num_rings,1))
            if '0,0,0,0' not in this_caption: success = True
        else:   # just skip this antinode
            print("\n\r",pad,":WARNING Can't fit an=",an,"\n",sep="",end="\r")
            this_caption = ""

        if (success):               # don't add blank lines, only add lines for success
            if (an > 0):
                caption+="\n"
            caption += this_caption
            boxes_arr.append(box)
    return img, caption



def handle_one_file(outdir, framenum):
    print("framenum = ",framenum)

    np_dims = (imHeight, imWidth, 1)     # for numpy, image dimensions are reversed
    img = 128*np.ones(np_dims, np.uint8)

    draw_waves(img)   # this is the main bottleneck, execution-time-wise

    max_antinodes = 6
    num_antinodes= random.randint(1,max_antinodes)
    img, caption = draw_antinodes(img, num_antinodes=num_antinodes)

    if (np.random.random() <= blur_prob): # blur image a bit
        blur_ksize = random.choice([3,5])
        img = blur_image(img, kernel_size=blur_ksize)

    # post-blur noise
    noise = cv2.randn(np.zeros(np_dims, np.uint8),40,40); # normal dist, mean 40 std 40
    img = cv2.add(img, noise)


    # further degrade image: drop some pixels
    mask = np.random.choice([0,1],size=img.shape).astype(np.float32)
    img = img*mask

    # finally replace background using real data
    img = bandpass_mixup(img)

    prefix = 'steelpan_'+str(framenum).zfill(7)
    cv2.imwrite(outdir+'/images/'+prefix+'.png',img)
    with open(outdir+'/annotations/'+prefix+meta_extension, "w") as text_file:
        text_file.write(caption+'\n')
    return


@call_parse
def gen_fake(
    n:Param("Number of images to generate", int)=2000,
    outdir:Param("Directory to write to",str)='espiownage-fake',
    ):
    "Generates fake ESPI-like images"

    random.seed(1)   # for reproducibility
    np.random.seed(1)

    mkdir_if_needed(outdir)
    mkdir_if_needed(outdir+'/images')
    mkdir_if_needed(outdir+'/annotations')

    parallel = True  # want this on, this one's slow
    if not parallel:
        for framenum in range(n):
            handle_one_file(outdir, framenum)
    else:
        wrapper = partial(handle_one_file, outdir)
        pool = mp.Pool(mp.cpu_count())
        results = pool.map(wrapper, range(n))
        pool.close()
        pool.join()
    return





# Cell
from .core import *
from .ellipse_editor import *
from .gen_bboxes import *
from .gen_crops import *
from .gen_masks import *
from .gen_fake import *
from .grab_recent import *

# Cell
from .core import *
from .ellipse_editor import *
from .gen_bboxes import *
from .gen_crops import *
from .gen_masks import *
from .gen_fake import *
from .grab_recent import *

# Cell
from .core import *
from .ellipse_editor import *
from .gen_bboxes import *
from .gen_crops import *
from .gen_masks import *
from .gen_fake import *
from .grab_recent import *